unit PtEvents;
//  Versions:     Delphi 6
//  Created by:   Peter Thnell
//  Contact:      thonellp@hotmail.com
//  When:         2002-08
//  Copyright:    Peter Thnell
//
//  Purpose of unit:
//    By using a TNotifyEvents property instead of a method property for events,
//    you enable an event to be broadcast to any number event methods.
//  Usage:
//    - Make sure that objects receiving events unsubscribe before they are
//    freed.  The multiplexer itself can be freed without notifications.
//    - You can make event multiplexers that handle other method types by
//    inheriting from TNotifyEvents and overriding CallEvent() and replacing
//    Subscribe() and Unsubscribe().

interface

uses
  Classes;

type
  TNotifyEvents = class
  private
    fEvents : array of TNotifyEvent;
    fCount  : integer;
    fEnabled: Boolean;
    fOwner: TObject;
    procedure SetCount(const Value: integer);
    function  IndexOf(const Event: TNotifyEvent): integer;
  protected
    property  Count : integer read fCount;
    function  IsSameMethod(const Event1, Event2: TNotifyEvent): Boolean;
    procedure CallEvent(const Event: TNotifyEvent); virtual;
  public
    constructor Create(const AOwner: TObject);
    property  Owner   : TObject read fOwner;
    property  Enabled : Boolean read fEnabled write fEnabled;
    procedure Call;
    procedure Subscribe(const Event: TNotifyEvent);
    procedure Unsubscribe(const Event: TNotifyEvent);
    procedure Clear;
  end;

implementation

type
  TMethod = record
    Method    : Pointer;  //  Pointer to function
    Instance  : Pointer;  //  Pointer to instance
  end;

{ TNotifyEvents }

procedure TNotifyEvents.Call;
var
  i : integer;
begin
  if  Enabled then
    for i := 0 to Count-1 do
      CallEvent( fEvents[i] );
end;

procedure TNotifyEvents.CallEvent(const Event: TNotifyEvent);
begin
  Event( Owner )
end;

procedure TNotifyEvents.Clear;
begin
  SetCount( 0 );
end;

constructor TNotifyEvents.Create(const AOwner: TObject);
begin
  fOwner    := AOwner;
  fEnabled  := true;
end;

function TNotifyEvents.IndexOf(const Event: TNotifyEvent): integer;
begin
  Result  := Count-1;
  while (Result >= 0) and not IsSameMethod( fEvents[Result], Event )  do
    dec( Result );
end;

function TNotifyEvents.IsSameMethod(const Event1, Event2: TNotifyEvent): Boolean;
begin
  Result  := (TMethod( Event1 ).Instance = TMethod( Event2 ).Instance) and
             (TMethod( Event1 ).Method   = TMethod( Event2 ).Method  )
end;

procedure TNotifyEvents.SetCount(const Value: integer);
var
  Capacity  : integer;
begin
  fCount    := Value;
  //  Determine and set the capacity required
  Capacity  := Length( fEvents );
  if  fCount = 0  then
    Capacity  := 0
  else if fCount < (Capacity shr 2)  then
    repeat
      Capacity  := Capacity shr 1;
    until fCount >= (Capacity shr 2)
  else
    begin
      if  Capacity = 0  then
        Capacity  := 4;
      while fCount > Capacity  do
        Capacity  := Capacity shl 1;
    end;
  SetLength( fEvents, Capacity );
end;

procedure TNotifyEvents.Subscribe(const Event: TNotifyEvent);
var
  i : integer;
begin
  //  Avoid duplicate entries
  i := IndexOf( Event );
  if  i < 0 then
    begin
      //  Add event
      SetCount( Count + 1 );
      fEvents[Count-1]  := Event;
    end;
end;

procedure TNotifyEvents.Unsubscribe(const Event: TNotifyEvent);
var
  i : integer;
begin
  //  Do not unsubscribe what isn't there
  i := IndexOf( Event );
  if  i >= 0  then
    begin
      //  Remove event
      fEvents[i]  := fEvents[Count-1];
      SetCount( Count - 1 );
    end;
end;

end.
